#Copyright (c) 2019-2021, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from typing import Optional, List
import numpy as np

from grid2op.dtypes import dt_int

from grid2op.Opponent import BaseOpponent
from grid2op.Opponent import GeometricOpponent
from grid2op.Exceptions import OpponentError


class GeometricOpponentMultiArea(BaseOpponent):
    """
    This opponent is a combination of several similar opponents (of Kind Geometric Opponent at this stage) attacking on different areas.
    The difference between unitary opponents is mainly the attackable lines (which belongs to different pre-identified areas
    """

    def __init__(self, action_space):
        BaseOpponent.__init__(self, action_space)
        self.list_opponents : Optional[List[GeometricOpponent]] = None
        self._new_attack_time_counters : Optional[np.ndarray] = None
        self._previous_attacks = None

    def init(
        self,
        partial_env,
        lines_attacked=None,
        attack_every_xxx_hour=24,
        average_attack_duration_hour=4,
        minimum_attack_duration_hour=2,
        pmax_pmin_ratio=4,
        **kwargs,
    ):
        """
        Generic function used to initialize the derived classes. For example, if an opponent reads from a file, the
        path where is the file is located should be pass with this method.
        This is based on init from GeometricOpponent, only parameter lines_attacked becomes a list of list

        Parameters
        ----------
        partial_env: grid2op Environment
            see the GeometricOpponent::init documentation

        lines_attacked: ``list(list)``
            The lists of lines attacked by each unitary opponent (this is a list of list: the size 
            of the outer list is the number of underlying opponent / number of areas and for each inner 
            list it gives the name of the lines to attack.)

        attack_every_xxx_hour: ``float``
            see the GeometricOpponent::init documentation

        average_attack_duration_hour: ``float``
            see the GeometricOpponent::init documentation

        minimum_attack_duration_hour: ``int``
            see the GeometricOpponent::init documentation

        pmax_pmin_ratio: ``float``
            see the GeometricOpponent::init documentation

        """

        if lines_attacked is None:
            partial_env.logger.warning("GeometricOpponentMultiArea: no area provided, the opponent will be deactivated.")
            return
        
        self.list_opponents = [GeometricOpponent(action_space=self.action_space) for _ in lines_attacked]
        self._previous_attacks = [None for _ in lines_attacked]

        for lines_attacked, opp in zip(lines_attacked, self.list_opponents):
            opp.init(
                partial_env=partial_env,
                lines_attacked=lines_attacked,
                attack_every_xxx_hour=attack_every_xxx_hour,
                average_attack_duration_hour=average_attack_duration_hour,
                minimum_attack_duration_hour=minimum_attack_duration_hour,
                pmax_pmin_ratio=pmax_pmin_ratio,
                **kwargs,
            )
        self._new_attack_time_counters = np.array([-1 for _ in lines_attacked])#ou plutôt 0 comme dans Geometric Opponent ?

    def reset(self, initial_budget):
        self._new_attack_time_counters = np.array([-1 for _ in self.list_opponents])

        for opp in self.list_opponents:  # maybe loop in different orders each time
            opp.reset(initial_budget)

    def attack(self, observation, agent_action, env_action, budget, previous_fails):
        """
        This method is the equivalent of "attack" for a regular agent.
        Opponent, in this framework can have more information than a regular agent (in particular it can
        view time step t+1), it has access to its current budget etc.
        Here we take the combination of unitary opponent attacks if they happen at the same time.
        We choose the attack duration as the minimum duration of several simultaneous attacks if that happen.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            see the GeometricOpponent::attack documentation
        opp_reward: ``float``
            see the GeometricOpponent::attack documentation
        done: ``bool``
            see the GeometricOpponent::attack documentation
        agent_action: :class:`grid2op.Action.Action`
            see the GeometricOpponent::attack documentation
        env_action: :class:`grid2op.Action.Action`
            see the GeometricOpponent::attack documentation
        budget: ``float``
            see the GeometricOpponent::attack documentation
        previous_fails: ``bool``
            see the GeometricOpponent::attack documentation
        Returns
        -------
        attack: :class:`grid2op.Action.Action`
            see the GeometricOpponent::attack documentation
        duration: ``int``
            see the GeometricOpponent::attack documentation
        """

        #go through opponents and check if attack or not. As soon as one attack, stop there
        self._new_attack_time_counters -= 1
        self._new_attack_time_counters[self._new_attack_time_counters < -1] = -1

        attack_combined = None
        for opp_id, opp in enumerate(self.list_opponents):
            if self._new_attack_time_counters[opp_id] == -1:
                attack_opp, attack_duration_opp = opp.attack(observation, agent_action, env_action, budget, previous_fails)
                if attack_opp is not None:
                    self._new_attack_time_counters[opp_id] = attack_duration_opp
                    self._previous_attacks[opp_id] = attack_opp
                    if attack_combined is None:
                        attack_combined = attack_opp.copy()
                    else:
                        attack_combined += attack_opp
                else:
                    self._previous_attacks[opp_id] = None
            else:
                opp.tell_attack_continues(observation, agent_action, env_action, budget)
                if attack_combined is None:
                    attack_combined = self._previous_attacks[opp_id].copy()
                else:
                    attack_combined += self._previous_attacks[opp_id]
        return attack_combined, 1


    def tell_attack_continues(self, observation, agent_action, env_action, budget):
        raise RuntimeError("I should not get there !")

    def get_state(self):
        return (self._new_attack_time_counters,
                self._previous_attacks,
                [opp.get_state() for opp in self.list_opponents])

    def set_state(self, my_state):
        self._new_attack_time_counters = np.array(my_state[0])
        self._previous_attacks = [el.copy() if el is not None else None for el in my_state[1]]
        for el, opp in zip(my_state[2], self.list_opponents):
            opp.set_state(el)
            
    def _custom_deepcopy_for_copy(self, new_obj, dict_=None):
        new_obj._new_attack_time_counters = 1 * self._new_attack_time_counters
        new_obj._previous_attacks = [el.copy() if el is not None else None 
                                     for el in self._previous_attacks]
        new_obj.list_opponents = []
        for opp in self.list_opponents:
            new_opp = type(opp).__new__(type(opp))
            opp._custom_deepcopy_for_copy(new_opp, dict_)
            new_obj.list_opponents.append(new_opp)
        super()._custom_deepcopy_for_copy(new_obj)
        return new_obj

    def seed(self, seed):
        """
        INTERNAL

         .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            We do not recommend to use this function outside of the two examples given in the description of this class.

        Set the seeds of the source of pseudo random number used for these several unitary opponents.

        Parameters
        ----------
        seed: ``int``
            The root seed to be set for the random number generator.

        Returns
        -------
        seeds: ``list``
            The associated list of seeds used.

        """
        seeds = []
        super().seed(seed)
        max_seed = np.iinfo(dt_int).max  # 2**32 - 1
        for opp in self.list_opponents:
            this_seed = self.space_prng.randint(max_seed)
            seeds.append(opp.seed(this_seed))
        return (seed, seeds)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
from grid2op.Action import PlayableAction
from grid2op.Observation import CompleteObservation

class ActionWCCI2022(PlayableAction):
    authorized_keys = {
        "set_line_status",
        "change_line_status",
        "set_bus",
        "change_bus",
        "redispatch",
        "set_storage",
        "curtail"
        }

    attr_list_vect = [
        "_set_line_status",
        "_switch_line_status",
        "_set_topo_vect",
        "_change_bus_vect",
        "_redispatch",
        "_storage_power",
        "_curtail"
        ]
    attr_list_set = set(attr_list_vect)
    pass
    
class ObservationWCCI2022(CompleteObservation):
    attr_list_vect = [
        "year",
        "month",
        "day",
        "hour_of_day",
        "minute_of_hour",
        "day_of_week",
        "gen_p",
        "gen_q",
        "gen_v",
        "load_p",
        "load_q",
        "load_v",
        "p_or",
        "q_or",
        "v_or",
        "a_or",
        "p_ex",
        "q_ex",
        "v_ex",
        "a_ex",
        "rho",
        "line_status",
        "timestep_overflow",
        "topo_vect",
        "time_before_cooldown_line",
        "time_before_cooldown_sub",
        "time_next_maintenance",
        "duration_next_maintenance",
        "target_dispatch",
        "actual_dispatch",
        "storage_charge",
        "storage_power_target",
        "storage_power",
        "gen_p_before_curtail",
        "curtailment",
        "curtailment_limit",
        "curtailment_limit_effective",  # starting grid2op version 1.6.6
        "_shunt_p",
        "_shunt_q",
        "_shunt_v",
        "_shunt_bus",  # starting from grid2op version 1.6.0
        "current_step",
        "max_step",  # starting from grid2op version 1.6.4
        "delta_time",  # starting grid2op version 1.6.5
        "gen_margin_up",
        "gen_margin_down",  # starting grid2op version 1.6.6
    ]
    attr_list_json = [
        "_thermal_limit",
        "support_theta",
        "theta_or",
        "theta_ex",
        "load_theta",
        "gen_theta",
        "storage_theta",
    ]
    attr_list_set = set(attr_list_vect)


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import warnings

from grid2op.Exceptions import Grid2OpException
from grid2op.Rules.BaseRules import BaseRules
from grid2op.Rules.AlwaysLegal import AlwaysLegal


class RulesChecker(object):
    """
    Class that define the rules of the game.

    """

    def __init__(self, legalActClass=AlwaysLegal):
        """

        Parameters
        ----------
        legalActClass: ``type``
            The class that will be used to tell if the actions are legal or not. The class must be given, and not
            an object of this class. It should derived from :class:`BaseRules`.
        """
        if isinstance(legalActClass, type):
            if not issubclass(legalActClass, BaseRules):
                raise Grid2OpException(
                    "Gamerules: legalActClass should be initialize with a class deriving "
                    "from BaseRules and not {}".format(type(legalActClass))
                )
            self.legal_action = legalActClass()
        else:
            if not isinstance(legalActClass, BaseRules):
                raise Grid2OpException(
                    'Parameter "legalActClass" used to build the Environment should be an instance of the '
                    'grid2op.BaseRules class, type provided is "{}"'.format(
                        type(legalActClass)
                    )
                )
            try:
                self.legal_action = copy.deepcopy(legalActClass)
            except Exception as exc_:
                warnings.warn("You passed the legal action as an instance that cannot be deepcopied. It will be "
                              "used 'as is', we do not garantee anything if you modify the original object.")
                self.legal_action = legalActClass
                
    def initialize(self, env):
        """
        This function is used to inform the class instance about the environment specification. 
        It can be the place to assert the defined rules are suited for the environement.
        Parameters
        ----------
        env: :class:`grid2op.Environment.Environment`
            The environment on which the action is performed. 
        """
        self.legal_action.initialize(env)

    def __call__(self, action, env):
        """
        Says if an action is legal or not.

        Parameters
        ----------
        action: :class:`grid2op.Action.Action`
            The action that need to be tested

        env: :class:`grid2op.Environment.Environment`
            The current used environment.

        Returns
        -------
        is_legal: ``bool``
            Assess if the given action is legal or not. ``True``: the action is legal, ``False`` otherwise
        reason:
            A grid2op IllegalException given the reason for which the action is illegal
        """
        return self.legal_action(action, env)



# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from typing import Tuple
import warnings

from grid2op.Exceptions import HandlerError
from grid2op.Chronics.handlers.baseHandler import BaseHandler


class PerfectForecastHandler(BaseHandler):    
    """This class is allows to generate "perfect forecast", with this class the agent
    will know what will be the exact production, loads etc for the near future.
    
    This is a strong "assumption" and it is not realistic. 
    
    To have make things more realistic, you can use the :class:`NoisyForecastHandler` but again, 
    this class is far from perfect.
    
    More "research" is needed in this area and any contribution is more than welcome !
    
    As the name suggest, you should use this class only for the FORECAST data and not for environment or maintenance.
    
    .. warning::
        It only works if the handlers of the environments supports the :func:`BaseHandler.get_future_data` is implemented
        for the environment handlers. 
        
    """
    def __init__(self, array_name, max_iter=-1, quiet_warnings : bool=False):
        super().__init__(array_name, max_iter)
        self.quiet_warnings = quiet_warnings
             
    def initialize(self, order_backend_arrays, names_chronics_to_backend):
        # nothing particular to do at initialization
        pass
    
    def done(self):
        # this handler is never "done", only when the "real data" it depends on is done
        return False
    
    def load_next(self, dict_):
        raise HandlerError("You should only use this class for FORECAST data, and not for ENVIRONMENT data. "
                           "You might want to use the `DoNothingHandler` that will `do_nothing` for "
                           "the environment data (see `from grid2op.Chronics.handlers import DoNothingHandler`)")
        
    def check_validity(self, backend):
        return True
    
    def load_next_maintenance(self):
        raise HandlerError("You should only use this class for FORECAST data and not for MAINTENANCE data")
    
    def load_next_hazard(self):
        raise HandlerError("You should only use this class for FORECAST data and not for HAZARDS data")

    def forecast(self,
                 forecast_horizon_id : int,
                 inj_dict_env : dict,
                 inj_dict_previous_forecast : dict,
                 # eg gen_p_handler if this is set to gen_p_for_handler:
                 env_handler : BaseHandler,  
                 # list of the 4 env handlers: (load_p_handler, load_q_handler, gen_p_handler, gen_v_handler)
                 env_handlers : Tuple[BaseHandler, BaseHandler, BaseHandler, BaseHandler]
                 ):
        res = env_handler.get_future_data(self._h_forecast[forecast_horizon_id], self.quiet_warnings)
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Observation.baseObservation import BaseObservation


class CompleteObservation(BaseObservation):
    """
    This class represent a complete observation, where everything on the powergrid can be observed without
    any noise.

    This is the only :class:`BaseObservation` implemented (and used) in Grid2Op. Other type of observation, for other
    usage can of course be implemented following this example.

    It has the same attributes as the :class:`BaseObservation` class. Only one is added here.

    For a :class:`CompleteObservation` the unique representation as a vector is:

        1. :attr:`BaseObservation.year` the year [1 element]
        2. :attr:`BaseObservation.month` the month [1 element]
        3. :attr:`BaseObservation.day` the day [1 element]
        4. :attr:`BaseObservation.hour_of_day` the hour of the day [1 element]
        5. :attr:`BaseObservation.minute_of_hour` minute of the hour  [1 element]
        6. :attr:`BaseObservation.day_of_week` the day of the week. Monday = 0, Sunday = 6 [1 element]
        7. :attr:`BaseObservation.gen_p` the active value of the productions
           [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        8. :attr:`BaseObservation.gen_q` the reactive value of the productions
           [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        9. :attr:`BaseObservation.gen_v` the voltage setpoint of the productions
           [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        10. :attr:`BaseObservation.load_p` the active value of the loads
            [:attr:`grid2op.Space.GridObjects.n_load` elements]
        11. :attr:`BaseObservation.load_q` the reactive value of the loads
            [:attr:`grid2op.Space.GridObjects.n_load` elements]
        12. :attr:`BaseObservation.load_v` the voltage setpoint of the loads
            [:attr:`grid2op.Space.GridObjects.n_load` elements]
        13. :attr:`BaseObservation.p_or` active flow at origin of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        14. :attr:`BaseObservation.q_or` reactive flow at origin of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        15. :attr:`BaseObservation.v_or` voltage at origin of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        16. :attr:`BaseObservation.a_or` current flow at origin of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        17. :attr:`BaseObservation.p_ex` active flow at extremity of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        18. :attr:`BaseObservation.q_ex` reactive flow at extremity of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        19. :attr:`BaseObservation.v_ex` voltage at extremity of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        20. :attr:`BaseObservation.a_ex` current flow at extremity of powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        21. :attr:`BaseObservation.rho` line capacity used (current flow / thermal limit)
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        22. :attr:`BaseObservation.line_status` line status [:attr:`grid2op.Space.GridObjects.n_line` elements]
        23. :attr:`BaseObservation.timestep_overflow` number of timestep since the powerline was on overflow
            (0 if the line is not on overflow)[:attr:`grid2op.Space.GridObjects.n_line` elements]
        24. :attr:`BaseObservation.topo_vect` representation as a vector of the topology [for each element
            it gives its bus]. See :func:`grid2op.Backend.Backend.get_topo_vect` for more information.
        25. :attr:`BaseObservation.time_before_cooldown_line` representation of the cooldown time on the powerlines
            [:attr:`grid2op.Space.GridObjects.n_line` elements]
        26. :attr:`BaseObservation.time_before_cooldown_sub` representation of the cooldown time on the substations
            [:attr:`grid2op.Space.GridObjects.n_sub` elements]
        27. :attr:`BaseObservation.time_next_maintenance` number of timestep before the next maintenance (-1 means
            no maintenance are planned, 0 a maintenance is in operation) [:attr:`grid2op.Space.GridObjects.n_line` elements]
        28. :attr:`BaseObservation.duration_next_maintenance` duration of the next maintenance. If a maintenance
            is taking place, this is the number of timestep before it ends. [:attr:`grid2op.Space.GridObjects.n_line` elements]
        29. :attr:`BaseObservation.target_dispatch` the target dispatch for each generator
            [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        30. :attr:`BaseObservation.actual_dispatch` the actual dispatch for each generator
            [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        31. :attr:`BaseObservation.storage_charge` the actual state of charge of each storage unit
            [:attr:`grid2op.Space.GridObjects.n_storage` elements]
        32. :attr:`BaseObservation.storage_power_target` the production / consumption of setpoint of each storage unit
            [:attr:`grid2op.Space.GridObjects.n_storage` elements]
        33. :attr:`BaseObservation.storage_power` the realized production / consumption of each storage unit
            [:attr:`grid2op.Space.GridObjects.n_storage` elements]
        34. :attr:`BaseObservation.gen_p_before_curtail` : the theoretical generation that would have happened
            if no generator from renewable energy sources have been performed (in MW)
            [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        35. :attr:`BaseObservation.curtailment` : the current curtailment applied
            [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        36. :attr:`BaseObservation.is_alarm_illegal` whether the last alarm has been illegal (due to budget
            constraint) [``bool``], 
            .. warning: /!\\\\ Only valid with "l2rpn_icaps_2021" environment /!\\\\ 
        37. :attr:`BaseObservation.curtailment_limit` : the current curtailment limit (if any)
            [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        38. :attr:`BaseObservation.time_since_last_alarm` number of step since the last alarm has been raised
            successfully [``int``]
            .. warning: /!\\\\ Only valid with "l2rpn_icaps_2021" environment /!\\\\ 
        39. :attr:`BaseObservation.last_alarm` : for each alarm zone, gives the last step at which an alarm has
            been successfully raised at this zone 
            .. warning: /!\\\\ Only valid with "l2rpn_icaps_2021" environment /!\\\\
            [:attr:`grid2op.Space.GridObjects.dim_alarms` elements]
        40. :attr:`BaseObservation.attention_budget` : the current attention budget
            [``int``]
        41. :attr:`BaseObservation.was_alarm_used_after_game_over` : was the last alarm used to compute anything related
            to the attention budget when there was a game over (can only be set to ``True`` if the observation
            corresponds to a game over), warning: /!\\\\ Only valid with "l2rpn_icaps_2021" environment /!\\\\ 
            [``bool``]
        42. :attr:`BaseObservation.is_alarm_illegal` whether the last alert has been illegal (due to budget
            constraint) [``bool``]
        43. :attr:`BaseObservation.curtailment_limit` : the current curtailment limit (if any)
            [:attr:`grid2op.Space.GridObjects.n_gen` elements]
        44. :attr:`BaseObservation.curtailment_limit_effective` Limit (in ratio of gen_pmax) imposed on 
            each renewable generator effectively imposed by the environment.
        45. :attr:`BaseObservation.current_step` the number of steps since the beginning of the episode (it's
            0 for the observation after a call to `env.reset()`)
        46. :attr:`BaseObservation.max_step` maximum number of steps that can be done by the environment.
            When :attr:`BaseObservation.current_step` is  :attr:`BaseObservation.max_step` the the environment
            is done.
        47. :attr:`BaseObservation.delta_time` Amount of time (in minutes) represented by a step. In general, there
            are the equivalent of 5 minutes between two steps.
        48. :attr:`BaseObservation.gen_margin_up` From how much can you increase each generators production between this
            step and the next.
        49. :attr:`BaseObservation.gen_margin_down` From how much can you decrease each generators production between this
            step and the next.   
        50. :attr:`BaseObservation.active_alert` This attribute gives the lines "under alert" at the given observation.
        51. :attr:`BaseObservation.time_since_last_alert`  Give the time since an alert has been raised for each powerline.
        52. :attr:`BaseObservation.alert_duration` Give the time since an alert has started for all attackable line.
        53. :attr:`BaseObservation.total_number_of_alert` Total number of alerts since the beginning of the episode sent by 
            the agent
        54. :attr:`BaseObservation.time_since_last_attack` For each attackable line `i` it counts the number of steps since the powerline has
            been attacked
        55. :attr:`BaseObservation.was_alert_used_after_attack` For each attackable line `i` it says if an alert has been used or not
            for the computation of the reward: +1 means "used and the alert was correct", -1 means "used and the alert was not correct"
            and 0 means "not used" 
        56. :attr:`BaseObservation.attack_under_alert` For each attackable line `i` it says if an alert has been sent (+1) or not (-1)
            for each attackable line currently under attack.
            
    """

    attr_list_vect = [
        "year",
        "month",
        "day",
        "hour_of_day",
        "minute_of_hour",
        "day_of_week",
        "gen_p",
        "gen_q",
        "gen_v",
        "load_p",
        "load_q",
        "load_v",
        "p_or",
        "q_or",
        "v_or",
        "a_or",
        "p_ex",
        "q_ex",
        "v_ex",
        "a_ex",
        "rho",
        "line_status",
        "timestep_overflow",
        "topo_vect",
        "time_before_cooldown_line",
        "time_before_cooldown_sub",
        "time_next_maintenance",
        "duration_next_maintenance",
        "target_dispatch",
        "actual_dispatch",
        "storage_charge",
        "storage_power_target",
        "storage_power",
        "gen_p_before_curtail",
        "curtailment",
        "curtailment_limit",
        "curtailment_limit_effective",  # starting grid2op version 1.6.6
        "is_alarm_illegal",
        "time_since_last_alarm",
        "last_alarm",
        "attention_budget",
        "was_alarm_used_after_game_over",
        "_shunt_p",
        "_shunt_q",
        "_shunt_v",
        "_shunt_bus",  # starting from grid2op version 1.6.0
        "current_step",
        "max_step",  # starting from grid2op version 1.6.4
        "delta_time",  # starting grid2op version 1.6.5
        "gen_margin_up",
        "gen_margin_down",  # starting grid2op version 1.6.6
        # line alert (starting grid2Op 1.9.1, for compatible envs)
        "active_alert",
        "attack_under_alert",
        "time_since_last_alert",
        "alert_duration",
        "total_number_of_alert",
        "time_since_last_attack",
        "was_alert_used_after_attack",
    ]
    attr_list_json = [
        "_thermal_limit",
        "support_theta",
        "theta_or",
        "theta_ex",
        "load_theta",
        "gen_theta",
        "storage_theta",
    ]
    attr_list_set = set(attr_list_vect)

    def __init__(self,
                 obs_env=None,
                 action_helper=None,
                 random_prng=None,
                 kwargs_env=None):

        BaseObservation.__init__(
            self,
            obs_env=obs_env,
            action_helper=action_helper,
            random_prng=random_prng,
            kwargs_env=kwargs_env
        )
        self._dictionnarized = None

    def update(self, env, with_forecast=True):
        # reset the matrices
        self._reset_matrices()
        self.reset()
        self._update_obs_complete(env, with_forecast=with_forecast)



__all__ = ["RandomObject", "SerializableSpace", "GridObjects"]

from grid2op.Space.RandomObject import RandomObject
from grid2op.Space.SerializableSpace import SerializableSpace
from grid2op.Space.GridObjects import GridObjects




# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

"""
Grid2Op

"""
__version__ = '1.9.3'

__all__ = [
    "Action",
    "Agent",
    "Backend",
    "Chronics",
    "Environment",
    "Exceptions",
    "Observation",
    "Parameters",
    "Rules",
    "Reward",
    "Runner",
    "Plot",
    "PlotGrid",
    "Episode",
    "Download",
    "VoltageControler",
    "tests",
    "main",
    "command_line",
    "utils",
    # utility functions
    "list_available_remote_env",
    "list_available_local_env",
    "get_current_local_dir",
    "change_local_dir",
    "list_available_test_env",
    "update_env",
    "make",
]


from grid2op.MakeEnv import  (make,
                              update_env,
                              list_available_remote_env,
                              list_available_local_env,
                              get_current_local_dir,
                              change_local_dir,
                              list_available_test_env
                             )


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Space import RandomObject


class BaseOpponent(RandomObject):
    def __init__(self, action_space):
        RandomObject.__init__(self)
        self.action_space = action_space
        self._do_nothing = self.action_space()

    def init(self, partial_env, **kwargs):
        """
        Generic function used to initialize the derived classes. For example, if an opponent reads from a file, the
        path where is the file is located should be pass with this method.
        """
        pass

    def reset(self, initial_budget):
        """
        This function is called at the end of an episode, when the episode is over. It aims at resetting the
        self and prepare it for a new episode.

        Parameters
        ----------
        initial_budget: ``float``
            The initial budget the opponent has
        """
        pass

    def attack(self, observation, agent_action, env_action, budget, previous_fails):
        """
        This method is the equivalent of "act" for a regular agent.

        Opponent, in this framework can have more information than a regular agent (in particular it can
        view time step t+1), it has access to its current budget etc.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The last observation (at time t)

        opp_reward: ``float``
            THe opponent "reward" (equivalent to the agent reward, but for the opponent) TODO do i add it back ???

        done: ``bool``
            Whether the game ended or not TODO do i add it back ???

        agent_action: :class:`grid2op.Action.Action`
            The action that the agent took

        env_action: :class:`grid2op.Action.Action`
            The modification that the environment will take.

        budget: ``float``
            The current remaining budget (if an action is above this budget, it will be replaced by a do nothing.

        previous_fails: ``bool``
            Wheter the previous attack failed (due to budget or ambiguous action)

        Returns
        -------
        attack: :class:`grid2op.Action.Action`
            The attack performed by the opponent. In this case, a do nothing, all the time.

        duration: ``int``
            The duration of the attack

        """
        # TODO maybe have a class "GymOpponent" where the observation would include the budget  and all other
        # TODO information, and forward something to the "act" method.
        return None, None

    def tell_attack_continues(self, observation, agent_action, env_action, budget):
        """
        The purpose of this method is to tell the agent that his attack is being continued
        and to indicate the current state of the grid.

        At every time step, either "attack" or "tell_acttack_continues" is called exactly once.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The last observation (at time t)

        agent_action: :class:`grid2op.Action.Action`
            The action that the agent took

        env_action: :class:`grid2op.Action.Action`
            The modification that the environment will take.

        budget: ``float``
            The current remaining budget (if an action is above this budget, it will be replaced by a do nothing.
        """
        pass

    def get_state(self):
        """
        This function should return the internal state of the Opponent.

        This means that after a call to `opponent.set_state(opponent.get_state())` the opponent should do the exact
        same things than without these calls.

        Returns
        -------

        """
        return None

    def set_state(self, my_state):
        """
        This function is used to set the internal state of the Opponent.

        Parameters
        ----------
        my_state

        """
        pass

    def _custom_deepcopy_for_copy(self, new_obj, dict_=None):
        super()._custom_deepcopy_for_copy(new_obj)
        if dict_ is None:
            dict_ = {}
        new_obj.action_space = self.action_space  # const
        new_obj._do_nothing = new_obj.action_space()
        new_obj.set_state(self.get_state())



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float


class L2RPNSandBoxScore(BaseReward):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            It **must not** serve as a reward. This scored needs to be **MINIMIZED**,
            and a reward needs to be maximized! Also, this "reward" is not scaled or anything. Use it as your
            own risk.

    Implemented as a reward to make it easier to use in the context of the L2RPN competitions, this "reward"
    computed the "grid operation cost". It should not be used to train an agent.

    The "reward" the closest to this score is given by the :class:`RedispReward` class.

    """

    def __init__(self,
                 alpha_redisp=1.0,
                 alpha_loss=1.0,
                 alpha_storage=1.0,
                 alpha_curtailment=1.0,
                 reward_max=1000.,
                 logger=None):
        BaseReward.__init__(self, logger=logger)
        self.reward_min = dt_float(1.0)  # carefull here between min and max...
        self.reward_max = dt_float(reward_max)
        self.alpha_redisp = dt_float(alpha_redisp)
        self.alpha_loss = dt_float(alpha_loss)
        self.alpha_storage = dt_float(alpha_storage)
        self.alpha_curtailment = dt_float(alpha_curtailment)

    def initialize(self, env):
        # TODO compute reward max! 
        return super().initialize(env)
    
    def _get_load_p(self, env):
        load_p, *_ = env.backend.loads_info()
        return load_p
    
    def _get_gen_p(self, env):
        gen_p, *_ = env.backend.generators_info()
        return gen_p
    
    def _get_losses(self, env, gen_p, load_p):
        return (gen_p.sum(dtype=dt_float) - load_p.sum(dtype=dt_float)) * env.delta_time_seconds / 3600.0
    
    def _get_marginal_cost(self, env):
        gen_activeprod_t = env._gen_activeprod_t
        p_t = np.max(env.gen_cost_per_MW[gen_activeprod_t > 0.0]).astype(dt_float)  
        # price is per MWh be sure to convert the MW (of losses and generation) to MWh before multiplying by the cost 
        return p_t
    
    def _get_redisp_cost(self, env, p_t):
        actual_dispatch = env._actual_dispatch
        c_redispatching = (
            np.abs(actual_dispatch).sum() * p_t * env.delta_time_seconds / 3600.0
        )
        return c_redispatching
    
    def _get_curtail_cost(self, env, p_t):
        curtailment_mw = -env._sum_curtailment_mw  # curtailment is always negative in the env 
        c_curtailment = (
            curtailment_mw * p_t * env.delta_time_seconds / 3600.0
        )
        return c_curtailment

    def _get_loss_cost(self, env, p_t):
        gen_p = self._get_gen_p(env)
        load_p = self._get_load_p(env)
        losses = self._get_losses(env, gen_p, load_p)
        c_loss = losses * p_t
        return c_loss
        
    def _get_storage_cost(self, env, p_t):
        c_storage = np.abs(env._storage_power).sum() * p_t * env.delta_time_seconds / 3600.0
        return c_storage
    
    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if has_error:
            # DO SOMETHING IN THIS CASE
            return self.reward_min

        # compute the marginal cost
        p_t = self._get_marginal_cost(env)
        
        # redispatching amount
        c_redispatching = self._get_redisp_cost(env, p_t)
        
        # curtailment amount
        c_curtailment = self._get_curtail_cost(env, p_t)
        
        # cost of losses
        c_loss = self._get_loss_cost(env, p_t)
        
        # storage units
        c_storage = self._get_storage_cost(env, p_t)
        
        # total "operationnal cost"
        c_operations = dt_float(self.alpha_loss * c_loss + 
                                self.alpha_redisp * c_redispatching + 
                                self.alpha_storage * c_storage + 
                                self.alpha_curtailment * c_curtailment)
        return c_operations



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from collections import OrderedDict
import numpy as np
import copy

from grid2op.dtypes import dt_int, dt_bool, dt_float
from grid2op.gym_compat.utils import check_gym_version, sample_seed
from grid2op.gym_compat.utils import GYM_AVAILABLE, GYMNASIUM_AVAILABLE


class __AuxBaseGymSpaceConverter:
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
        Used as a base class to convert grid2op state to gym state (wrapper for some useful function
        for both the action space and the observation space).

    .. warning::
        Depending on the presence absence of gymnasium and gym packages this class might behave differently.
        
        In grid2op we tried to maintain compatibility both with gymnasium (newest) and gym (legacy, 
        no more maintained) RL packages. The behaviour is the following:
        
        - :class:`_BaseGymSpaceConverter` will inherit from gymnasium if it's installed 
          (in this case it will be :class:`_BaseGymnasiumSpaceConverter`), otherwise it will
          inherit from gym (and will be exactly :class:`_BaseLegacyGymSpaceConverter`)
        - :class:`_BaseGymnasiumSpaceConverter` will inherit from gymnasium if it's available and never from
          from gym
        - :class:`_BaseLegacyGymSpaceConverter` will inherit from gym if it's available and never from
          from gymnasium
        
        See :ref:`gymnasium_gym` for more information
        
    """

    def __init__(self, dict_gym_space, dict_variables=None):
        check_gym_version(type(self)._gymnasium)
        type(self)._DictType.__init__(self, dict_gym_space)
        self._keys_encoding = {}
        if dict_variables is not None:
            for k, v in dict_variables.items():
                self._keys_encoding[k] = v
        self.__func = {}

    @classmethod
    def _generic_gym_space(cls, dt, sh, low=None, high=None):
        if dt == dt_int:
            if low is None:
                low = np.iinfo(dt).min
            if high is None:
                high = np.iinfo(dt).max
        else:
            low = -np.inf
            high = +np.inf
        shape = (sh,)
        my_type = cls._BoxType(
            low=dt.type(low), high=dt.type(high), shape=shape, dtype=dt
        )
        return my_type

    @classmethod
    def _boolean_type(cls, sh):
        return cls._MultiBinaryType(n=sh)

    @staticmethod
    def _simplifykeys_for_timestamps(key):
        """some keys are encoded to be returned as scalar, i need to transform them."""
        res = (
            (key == "year")
            or (key == "month")
            or (key == "day")
            or (key == "hour_of_day")
            or (key == "minute_of_hour")
            or (key == "day_of_week")
        )
        res = (
            res
            or (key == "is_alarm_illegal")
            or (key == "was_alarm_used_after_game_over")
            or (key == "is_alert_illegal")
            or (key == "was_alert_used_after_attack")
        )
        return res

    @staticmethod
    def _extract_obj_grid2op(vect, dtype, key):
        if len(vect) == 1 and _BaseGymSpaceConverter._simplifykeys_for_timestamps(key):
            res = vect[0]
            # convert the types for json serializable
            # this is not automatically done by gym...
            if dtype == dt_int or dtype == np.int64 or dtype == np.int32:
                res = int(res)
            elif dtype == dt_float or dtype == np.float64 or dtype == np.float32:
                res = float(res)
            elif dtype == dt_bool:
                res = bool(res)
        else:
            res = vect
        return res

    def _base_to_gym(self, keys, obj, dtypes, converter=None):
        """convert the obj (grid2op object) into a gym observation / action space"""
        res = OrderedDict()
        for k in keys:
            if k in self.__func:
                obj_json_cleaned = self.__func[k](obj)
            else:
                if converter is not None:
                    # for converting the names between internal names and "human readable names"
                    conv_k = converter[k]
                else:
                    conv_k = k

                obj_raw = obj._get_array_from_attr_name(conv_k)
                if conv_k in self._keys_encoding:
                    if self._keys_encoding[conv_k] is None:
                        # keys is deactivated
                        continue
                    elif isinstance(self._keys_encoding[conv_k], type(self)._SpaceType):
                        obj_json_cleaned = getattr(obj, conv_k)
                    else:
                        # i need to process the "function" part in the keys
                        obj_json_cleaned = self._keys_encoding[conv_k].g2op_to_gym(
                            obj_raw
                        )
                else:
                    obj_json_cleaned = self._extract_obj_grid2op(obj_raw, dtypes[k], k)
            res[k] = obj_json_cleaned
        return res

    def add_key(self, key_name, function, return_type):
        """

        Allows to add arbitrary function to the representation, as a gym environment of
        the action space of the observation space.

        TODO
        **NB** this key is not used when converted back to grid2Op object, as of now we don't recommend to
        use it for the action space !

        See the example for more information.

        Parameters
        ----------
        key_name:
            The name you want to get

        function:
            A function that takes as input

        return_type

        Returns
        -------


        Examples
        ---------
        In the example below, we explain how to add the "connectivity_matrix" as part of the observation space
        (when converted to gym). The new key "connectivity matrix" will be added to the gym observation.

        .. code-block:: python


            # create a grid2op environment
            import grid2op
            env_name = "l2rpn_case14_sandbox"
            env_glop = grid2op.make(env_name)

            # convert it to gym
            import gym
            import numpy as np
            from grid2op.gym_compat import GymEnv
            env_gym = GymEnv(env_glop)

            # default gym environment, the connectivity matrix is not computed
            obs_gym = env_gym.reset()
            print(f"Is the connectivity matrix part of the observation in gym: {'connectivity_matrix' in obs_gym}")

            # add the "connectivity matrix" as part of the observation in gym
            from gym.spaces import Box
            shape_ = (env_glop.dim_topo, env_glop.dim_topo)
            env_gym.observation_space.add_key("connectivity_matrix",
                                              lambda obs: obs.connectivity_matrix(),
                                              Box(shape=shape_,
                                                  low=np.zeros(shape_),
                                                  high=np.ones(shape_),
                                                )
                                              )

            # we highly recommend to "reset" the environment after setting up the observation space

            obs_gym = env_gym.reset()
            print(f"Is the connectivity matrix part of the observation in gym: {'connectivity_matrix' in obs_gym}")
        """

        self.spaces[key_name] = return_type
        self.__func[key_name] = function

    def get_dict_encoding(self):
        """
        TODO examples and description

        Returns
        -------

        """
        return copy.deepcopy(self._keys_encoding)

    def reencode_space(self, key, func):
        """
        TODO examples and description

        Returns
        -------

        """
        raise NotImplementedError(
            "This should be implemented in the GymActionSpace and GymObservationSpace"
        )

    def reenc(self, key, fun):
        """
        shorthand for :func:`GymObservationSpace.reencode_space` or
        :func:`GymActionSpace.reencode_space`
        """
        return self.reencode_space(key, fun)

    def keep_only_attr(self, attr_names):
        """
        keep only a certain part of the observation
        """
        if isinstance(attr_names, str):
            attr_names = [attr_names]

        dict_ = self.spaces.keys()
        res = self
        for k in dict_:
            if k not in attr_names:
                res = res.reencode_space(k, None)
        return res

    def ignore_attr(self, attr_names):
        """
        ignore some attribute names from the space
        """
        if isinstance(attr_names, str):
            attr_names = [attr_names]
        res = self
        for k in self.spaces.keys():
            if k in attr_names:
                res = res.reencode_space(k, None)
        return res

    def seed(self, seed=None):
        """Seed the PRNG of this space.
        see issue https://github.com/openai/gym/issues/2166
        of openAI gym
        """
        seeds = super(type(self)._DictType, self).seed(seed)
        sub_seeds = seeds
        max_ = np.iinfo(dt_int).max
        for i, space_key in enumerate(sorted(self.spaces.keys())):
            sub_seed = sample_seed(max_, self.np_random)
            sub_seeds.append(self.spaces[space_key].seed(sub_seed))
        return sub_seeds

    def close(self):
        pass


if GYM_AVAILABLE:
    from gym.spaces import (Discrete as LegGymDiscrete,
                            Box as LegGymBox,
                            Dict as LegGymDict,
                            Space as LegGymSpace,
                            MultiBinary as LegGymMultiBinary,
                            Tuple as LegGymTuple)
    _BaseLegacyGymSpaceConverter = type("_BaseLegacyGymSpaceConverter",
                                        (__AuxBaseGymSpaceConverter, LegGymDict, ),
                                        {"_DiscreteType": LegGymDiscrete,
                                         "_BoxType": LegGymBox,
                                         "_DictType": LegGymDict,
                                         "_SpaceType": LegGymSpace, 
                                         "_MultiBinaryType": LegGymMultiBinary, 
                                         "_TupleType": LegGymTuple,
                                         "_gymnasium": False,
                                         "__module__": __name__})
    _BaseLegacyGymSpaceConverter.__doc__ = __AuxBaseGymSpaceConverter.__doc__
    _BaseGymSpaceConverter = _BaseLegacyGymSpaceConverter
    _BaseGymSpaceConverter.__doc__ = __AuxBaseGymSpaceConverter.__doc__
        

if GYMNASIUM_AVAILABLE:
    from gymnasium.spaces import Discrete, Box, Dict, Space, MultiBinary, Tuple
    _BaseGymnasiumSpaceConverter = type("_BaseGymnasiumSpaceConverter",
                                        (__AuxBaseGymSpaceConverter, Dict, ),
                                        {"_DiscreteType": Discrete,
                                         "_BoxType": Box,
                                         "_DictType": Dict,
                                         "_SpaceType": Space, 
                                         "_MultiBinaryType": MultiBinary, 
                                         "_TupleType": Tuple, 
                                         "_gymnasium": True,
                                         "__module__": __name__})
    _BaseGymnasiumSpaceConverter.__doc__ = __AuxBaseGymSpaceConverter.__doc__
    _BaseGymSpaceConverter = _BaseGymnasiumSpaceConverter
    _BaseGymSpaceConverter.__doc__ = __AuxBaseGymSpaceConverter.__doc__



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Action import PlayableAction
from grid2op.Observation import CompleteObservation


class ActionICAPS2021(PlayableAction):
    authorized_keys = {
        "set_line_status",
        "change_line_status",
        "set_bus",
        "change_bus",
        "redispatch",
        "curtail",
        "raise_alarm",
        }

    attr_list_vect = ['_set_line_status',
                      '_switch_line_status',
                      '_set_topo_vect',
                      '_change_bus_vect', 
                      '_redispatch',
                      '_storage_power',
                      '_curtail',
                      '_raise_alarm']
    attr_list_set = set(attr_list_vect)
    pass


class ObservationICAPS2021(CompleteObservation):
    attr_list_vect = ['year',
                      'month',
                      'day',
                      'hour_of_day',
                      'minute_of_hour',
                      'day_of_week',
                      'gen_p',
                      'gen_q',
                      'gen_v',
                      'load_p',
                      'load_q',
                      'load_v',
                      'p_or',
                      'q_or',
                      'v_or',
                      'a_or',
                      'p_ex',
                      'q_ex',
                      'v_ex',
                      'a_ex',
                      'rho',
                      'line_status',
                      'timestep_overflow',
                      'topo_vect',
                      'time_before_cooldown_line',
                      'time_before_cooldown_sub',
                      'time_next_maintenance',
                      'duration_next_maintenance',
                      'target_dispatch',
                      'actual_dispatch',
                      'storage_charge',
                      'storage_power_target',
                      'storage_power',
                      'gen_p_before_curtail',
                      'curtailment',
                      'curtailment_limit',
                      'is_alarm_illegal',
                      'time_since_last_alarm',
                      'last_alarm',
                      'attention_budget',
                      'was_alarm_used_after_game_over',
                      '_shunt_p',
                      '_shunt_q',
                      '_shunt_v',
                      '_shunt_bus'
    ]
    
    attr_list_json = [
        "current_step",
        "max_step",
        "delta_time",
        "gen_margin_up",
        "gen_margin_down",
        "_thermal_limit",
        "support_theta",
        "theta_or",
        "theta_ex",
        "load_theta",
        "gen_theta",
        "storage_theta",
    ]
    attr_list_set = set(attr_list_vect)
    


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Action.playableAction import PlayableAction


class DontAct(PlayableAction):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        This type of action is only compatible with "do nothing"...

    This class is model the action where you force someone to do absolutely nothing. It is not the "do nothing"
    action.

    This is not the "do nothing" action. This class will not implement any way to modify the grid. Any attempt to
    modify it will fail.

    """

    authorized_keys = set()
    attr_list_vect = []

    def __init__(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        See the definition of :func:`BaseAction.__init__` and of :class:`BaseAction` for more information. Nothing
        more is done in this constructor.

        """
        PlayableAction.__init__(self)

    def update(self, dict_):
        """
        It has the particularity to not use `dict_`

        Parameters
        ----------
        dict_: ``dict``
            A dictionnary, not taken into account for this specific class.

        Returns
        -------

        """
        return self



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import warnings
import copy

from grid2op.Action.baseAction import BaseAction
from grid2op.Action.serializableActionSpace import SerializableActionSpace


class ActionSpace(SerializableActionSpace):
    """
    :class:`ActionSpace` should be created by an :class:`grid2op.Environment.Environment`
    with its parameters coming from a properly
    set up :class:`grid2op.Backend.Backend` (ie a Backend instance with a loaded powergrid.
    See :func:`grid2op.Backend.Backend.load_grid` for
    more information).

    It will allow, thanks to its :func:`ActionSpace.__call__` method to create valid :class:`BaseAction`. It is the
    the preferred way to create an object of class :class:`BaseAction` in this package.

    On the contrary to the :class:`BaseAction`, it is NOT recommended to overload this helper. If more flexibility is
    needed on the type of :class:`BaseAction` created, it is recommended to pass a different "*actionClass*" argument
    when it's built. Note that it's mandatory that the class used in the "*actionClass*" argument derived from the
    :class:`BaseAction`.

    Attributes
    ----------
    legal_action: :class:`grid2op.RulesChecker.BaseRules`
        Class specifying the rules of the game used to check the legality of the actions.

    """

    def __init__(
        self,
        gridobj,
        legal_action,
        actionClass=BaseAction,  # need to be a base grid2op type (and not a type generated on the fly)
    ):
        """
        INTERNAL USE ONLY

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            The actions space is created by the environment. Do not attempt to create one yourself.

        All parameters (name_gen, name_load, name_line, sub_info, etc.) are used to fill the attributes having the
        same name. See :class:`ActionSpace` for more information.

        Parameters
        ----------

        gridobj: :class:`grid2op.Space.GridObjects`
            The representation of the powergrid.

        actionClass: ``type``
            Note that this parameter expected a class and not an object of the class. It is used to return the
            appropriate action type.

        legal_action: :class:`grid2op.RulesChecker.BaseRules`
            Class specifying the rules of the game used to check the legality of the actions.

        """
        actionClass._add_shunt_data()
        actionClass._update_value_set()
        SerializableActionSpace.__init__(self, gridobj, actionClass=actionClass)
        self.legal_action = legal_action

    def __call__(
        self, dict_: dict = None, check_legal: bool = False, env: "BaseEnv" = None
    ) -> BaseAction:
        """
        This utility allows you to build a valid action, with the proper sizes if you provide it with a valid
        dictionary.

        More information about this dictionary can be found in the :func:`Action.update` help. This dictionary
        is not changed in this method.

        **NB** This is the only recommended way to make a valid, with proper dimension :class:`Action` object:

        Examples
        --------
        Here is a short example on how to make a action. For more detailed examples see :func:`Action.update`

        .. code-block:: python

            import grid2op
            # create a simple environment
            env = grid2op.make()
            act = env.action_space({})
            # act is now the "do nothing" action, that doesn't modify the grid.

        Parameters
        ----------
        dict_ : ``dict``
            see :func:`Action.__call__` documentation for an extensive help about this parameter

        check_legal: ``bool``
            is there a test performed on the legality of the action. **NB** When an object of class :class:`Action` is
            used, it is automatically tested for ambiguity. If this parameter is set to ``True`` then a legality test
            is performed. An action can be illegal if the environment doesn't allow it, for example if an agent tries
            to reconnect a powerline during a maintenance.

        env: :class:`grid2op.Environment.Environment`, optional
            An environment used to perform a legality check.

        Returns
        -------
        res: :class:`BaseAction`
            An action that is valid and corresponds to what the agent want to do with the formalism defined in
            see :func:`Action.udpate`.

        """

        res = self.actionClass()
        # update the action
        res.update(dict_)
        if check_legal:
            is_legal, reason = self._is_legal(res, env)
            if not is_legal:
                raise reason

        return res

    def _is_legal(self, action, env):
        """
        INTERNAL USE ONLY

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            Whether an action is legal or not is checked by the environment at each call
            to `env.step`

        Parameters
        ----------
        action: :class:`BaseAction`
            The action to test

        env: :class:`grid2op.Environment.Environment`
            The current environment

        Returns
        -------
        res: ``bool``
            ``True`` if the action is legal, ie is allowed to be performed by the rules of the game. ``False``
            otherwise.
        """
        if env is None:
            warnings.warn(
                "Cannot performed legality check because no environment is provided."
            )
            return True, None
        is_legal, reason = self.legal_action(action, env)
        return is_legal, reason

    def _custom_deepcopy_for_copy(self, new_obj):
        """implements a faster "res = copy.deepcopy(self)" to use
        in "self.copy"
        Do not use it anywhere else...
        """
        # TODO clean that after it is working... (ie make this method per class...)
        # fill the super classes
        super()._custom_deepcopy_for_copy(new_obj)

        # now fill my class
        new_obj.legal_action = copy.deepcopy(self.legal_action)

    def copy(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Perform a deep copy of the Observation space.

        """
        # performs the copy
        # res = copy.deepcopy(self)  # painfully slow...
        # create an empty "me"
        my_cls = type(self)
        res = my_cls.__new__(my_cls)
        self._custom_deepcopy_for_copy(res)
        return res

    def close(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Make sure all references to possible backends are closed. This is not used here in general but might be for some specific cases.

        """
        pass
